-----------------------------------------------------------------------------
-- |
-- Module      :  Conjugacy
-- Project     :  HBC
-- Copyright   :  (c) Hal Daume III
-- License     :  Open
-- 
-- Maintainer  :  Hal Daume III <me@hal3.name>
-- Stability   :  stable
-- Portability :  portable
--
-- Implementation of conjugacy simplification
--
-----------------------------------------------------------------------------

module Conjugacy where

import Core

import Data.List
import Data.Maybe
import Control.Monad
import Debug.Trace

-- to allow for user-defined conjugacy rules
data Conjugate = Conjugate { prior :: Id
                           , likelihood :: Id
                           , posterior  :: Id   -- allow distribution switching, eg., DirSym -> Dir
                           , matching   :: Int  -- which expression index in the posterior has to match the prior parameter
                           , suffstats  :: [Expr] -> Lik -> [Expr]
                           }

-- type of likelihoods
type Lik = [([Range],Dist,[Expr])]
          -- means prod_(rl,d,el) prod_rl d^(prod el)

-- standard rules
conjugacyRules =
    [ Conjugate { prior      = "Dir"
                , likelihood = "Mult"
                , posterior  = "Dir"
                , matching   = 0
                , suffstats  = conjDirMult
                }
    , Conjugate { prior      = "DirSym"
                , likelihood = "Mult"
                , posterior  = "Dir"
                , matching   = 0
                , suffstats  = conjDirSymMult
                }
    , Conjugate { prior      = "NorMV"
                , likelihood = "NorMV"
                , posterior  = "NorMV"
                , matching   = 0
                , suffstats  = conjNorNor True
                }
    , Conjugate { prior      = "Nor"
                , likelihood = "Nor"
                , posterior  = "Nor"
                , matching   = 0
                , suffstats  = conjNorNor False
                }
    , Conjugate { prior      = "Gam"
                , likelihood = "Exp"
                , posterior  = "Gam"
                , matching   = 0
                , suffstats  = conjGamExp
                }
    , Conjugate { prior      = "Bet"
                , likelihood = "Bin"
                , posterior  = "Bet"
                , matching   = 0
                , suffstats  = conjBetBin
                }
    , Conjugate { prior      = "Gam"
                , likelihood = "DirSym"
                , posterior  = "Gam"
                , matching   = 0
                , suffstats  = conjGamDirSym
                }
    , Conjugate { prior      = "Gam"
                , likelihood = "Poi"
                , posterior  = "Gam"
                , matching   = 0
                , suffstats  = conjGamPoi
                }
    , Conjugate { prior      = "IG"
                , likelihood = "NorMV"
                , posterior  = "IG"
                , matching   = 1
                , suffstats  = conjGamNor True
                }
    , Conjugate { prior      = "IG"
                , likelihood = "Nor"
                , posterior  = "IG"
                , matching   = 1
                , suffstats  = conjGamNor False
                }
    ]


conjDirMult [al,len] lik = [foldr conj al lik, len]
  where
    conj (rl, Dist v _ _, el) d0 = Func "+" [d0, foldr makeSum (foldr makeProd' v0 el) rl]
      where v0 = Func "ID" [Var v, len]
    
conjDirSymMult [al,len] lik = [foldr conj al0 lik, len]
  where
    al0 = Func "vec" [al, Con (I 1), len]
    conj (rl, Dist v _ _, el) d0 = Func "+" [d0, foldr makeSum (foldr makeProd' v0 el) rl]
      where v0 = Func "ID" [Var v, len]

conjGamDirSym [a,b] lik = [a, b']
  where
--    b'    = Func "/" [b, Func "-" [one, Func "*" [b, foldr conj (Con (R 0)) lik]]]
    b' = Func "-" [Func "/" [one, b], foldr conj (Con (R 0)) lik]
    one   = Con (R 1)
    conj (rl, Dist v _ [_,dim], el) d0 = 
      Func "+" [d0, foldr makeLogSum (foldr makeProd (sumlog dim v) el) rl]
    makeLogSum r e = SP False r e ""
    sumlog dim (V v vl) = SP False (Range "cgds" (Con (I 1)) dim) (Func "log" [Var (V v (vl ++ [Id "cgds" Nothing]))]) ""
--    sumlog e = Func "sum" [Func "log" [e]]

makeSum  r e  = SP False r e ""
makeProd  e0 e = Func  "*" [e0, e]
makeProd' e0 e = Func ".*" [e0, e]

conjGamNor isMultivar [al,be] lik = [al', be']
  where
    al' = Func "+" [al, Func "/" [n   , Con (R 2)]]
    be' = Func "+" [be, Func "/" [sumx, Con (R 2)]]

    sumx = foldr1 (\a b -> Func "+" [a,b]) $ map (makePost True ) lik
    n    = foldr1 (\a b -> Func "+" [a,b]) $ map (makePost False) lik

    one = Con (R 1)

    makePost True  (rl, Dist v _ (mu:_), el) = foldr makeSum (foldr makeProd' (d mu $ Var v) el) rl
    makePost False (rl, Dist v _ _     , el) = foldr makeSum (foldr makeProd  one            el) rl

    d mu e 
      | isMultivar = Func "sqrDiff" [e,mu]
      | otherwise  = Func "*" [Func "-" [e,mu], Func "-" [e,mu]]
--      | isMultivar = Func "sum" [Func "*" [Func "-" [e,mu], Func "-" [e,mu]]]
--      | otherwise  = Func "*" [Func "-" [e,mu], Func "-" [e,mu]]


conjNorNor isMultivar (mu0:si0:rest) lik 
  | length all_si /= 1 = error ("Conjugacy.conjNorNor: not all likelihood variances are the same... this is fixable, but not yet implemented")
  | length all_mu /= 1 = error ("Conjugacy.conjNorNor: not all likelihood means are the same... this is fixable, but not yet implemented")
  | otherwise = (mu' : rho' : rest)
  where
    all_si = nub [si | (_,Dist _ _ (_:si:_),_) <- lik]
    all_mu = nub [si | (_,Dist _ _ (mu:_:_),_) <- lik]
    si     = head all_si

    --   = (1/si0 + n/si)^{-1}; n = sum_...
    rho' = Func "^" [Func "+" [Func "/" [one, si0]
                              ,Func "*" [Func "/" [one, si], n]]
                    ,Con (R (-1))]

    times = if isMultivar then ".*" else "*"

    --   = rho' * (mu0/si0 + (1/si) sum x)
    mu'  = Func times [rho'
                      ,Func "+" [Func times [Func "/" [one, si0], mu0]
                                ,Func times [Func "/" [one, si ], sumx]]]

    sumx = foldr1 (\a b -> Func "+" [a,b]) $ map (makePost True ) lik
    n    = foldr1 (\a b -> Func "+" [a,b]) $ map (makePost False) lik

    -- helpers
    one = Con (R 1)

    makePost True  (rl, Dist v _ _, el) = foldr makeSum (foldr makeProd' (Var v) el) rl
    makePost False (rl, Dist v _ _, el) = foldr makeSum (foldr makeProd  one     el) rl

conjGamExp [al,be] lik = [al',be']
  where
    al' = Func "+" [al, n]
    be' = Func "/" [be, Func "+" [one, Func "*" [be, sumx]]]

    sumx = foldr1 (\a b -> Func "+" [a,b]) $ map (makePost True ) lik
    n    = foldr1 (\a b -> Func "+" [a,b]) $ map (makePost False) lik

    one = Con (R 1)

    makePost True  (rl, Dist v _ _, el) = foldr makeSum (foldr makeProd' (Var v) el) rl
    makePost False (rl, Dist v _ _, el) = foldr makeSum (foldr makeProd  one     el) rl

conjGamPoi [al,be] lik = [al', be']
  where
    al' = Func "+" [al, sumx]
    be' = Func "/" [be, Func "+" [one, Func "*" [be, n]]]

    sumx = foldr1 (\a b -> Func "+" [a,b]) $ map (makePost True ) lik
    n    = foldr1 (\a b -> Func "+" [a,b]) $ map (makePost False) lik

    one = Con (R 1)

    makePost True  (rl, Dist v _ _, el) = foldr makeSum (foldr makeProd' (Var v) el) rl
    makePost False (rl, Dist v _ _, el) = foldr makeSum (foldr makeProd  one     el) rl


conjBetBin [al,be] lik = [foldr (conj True ) al lik
                         ,foldr (conj False) al lik]
  where
    conj b (rl, Dist v _ _, el) d0 = Func "+" [d0, foldr makeSum (foldr makeProd (Con (I 1)) (if not b then map notit el else el)) rl]
    notit e = Func "not" [e]

simplify :: Core -> Core
simplify (Core l) = Core [(v,applyConjugacy conjugacyRules d,r) | (v,d,r) <- l]

applyConjugacy :: [Conjugate] -> Dist -> Dist
applyConjugacy rules d 
  | null pr_lik = d
  | otherwise   = applyConjugacy rules post
  where
    pr_lik = [ (prior, (lik,rest))
                   | prior <- findPrior conjugacyRules d
                   , let (lik,rest) = findLikelihood prior d
                   , not (null lik)
                   , isJust rest
             ]
    (pr,lik) = head pr_lik
    post     = computePosterior pr lik

findPrior :: [Conjugate] -> Dist ->  [(Dist, Conjugate)]
findPrior rules d@(Dist _ id _) = [(d,r) | r <- rules, prior r == id]
findPrior rules (a :*: b) = findPrior rules a `mplus`  findPrior rules b
findPrior rules (a :^: _) = findPrior rules a
findPrior rules (Prod _ d _) = findPrior rules d

findLikelihood :: (Dist, Conjugate) -> Dist -> (Lik, Maybe Dist) -- second is what remains
findLikelihood (Dist priorParam _ _, conj) d@(Dist _ id el)
  | id == likelihood conj &&
    matching conj < length el =
    case matchVariable priorParam (el !! matching conj) of
      Nothing  -> ([], Just d)
      Just el' -> ([([],d,el')], Nothing)
  | otherwise = ([], Just d)
findLikelihood dc (a :*: b) = (likA++likB, restA `distTimes` restB)
  where (likA,restA) = findLikelihood dc a
        (likB,restB) = findLikelihood dc b
findLikelihood dc (a :^: e) = (map (\ (r,d,el) -> (r,d,e:el)) lik, rest)
  where (lik,rest) = findLikelihood dc a
findLikelihood dc (Prod range d _) = (map (\ (r,d,el) -> (range:r,d,el)) lik, rest)
  where (lik,rest) = findLikelihood dc d

matchVariable v e | Var v == e = Just []
matchVariable (V pid pse) (Var (V eid ese))
  | pid == eid && length pse == length ese = Just (zipWith (\p e -> Func "=" [sexprToExpr p,sexprToExpr e]) pse ese)
matchVariable _ _ = Nothing

computePosterior :: (Dist, Conjugate) -> (Lik, Maybe Dist) -> Dist
computePosterior _ (_,Nothing) = error ("Conjugacy.computePosterior: Maybe Dist is Nothing; at least the prior should be there")
computePosterior (prior@(Dist id0 _ el0), conj) (lik, Just rest) = 
  fromJust ((rest `distWithout` prior) `distTimes` Just (Dist id0 (posterior conj) ((suffstats conj) el0 lik)))
